from bsb.unittest import FixedPosConfigFixture, RandomStorageFixture, NumpyTestCase
from bsb.core import Scaffold
from bsb_hdf5.connectivity_set import LocationOutOfBoundsError
import unittest


class TestConnectivitySet(
    FixedPosConfigFixture,
    RandomStorageFixture,
    NumpyTestCase,
    unittest.TestCase,
    engine_name="hdf5",
):
    def setUp(self):
        super().setUp()
        self.cfg.connectivity.add(
            "all_to_all",
            dict(
                strategy="bsb.connectivity.AllToAll",
                presynaptic=dict(cell_types=["test_cell"]),
                postsynaptic=dict(cell_types=["test_cell"]),
            ),
        )
        self.network = Scaffold(self.cfg, self.storage)
        self.network.compile(clear=True, skip_connectivity=True)

    def test_pre_oob(self):
        f = self.network.connectivity.all_to_all.connect_cells

        def pre_oob_connect(pre_set, post_set, src_locs, dest_locs, tag=None):
            f(pre_set, post_set, src_locs + [100, 0, 0], dest_locs + [100, 0, 0])

        self.network.connectivity.all_to_all.connect_cells = pre_oob_connect
        with self.assertRaises(LocationOutOfBoundsError):
            self.network.compile(append=True, skip_placement=True)



from bsb import config
from bsb.services import MPILock
from bsb.storage.interfaces import Engine, StorageNode as IStorageNode, NoopLock
from .placement_set import PlacementSet
from .connectivity_set import ConnectivitySet
from .file_store import FileStore
from .morphology_repository import MorphologyRepository
from datetime import datetime
import json
import h5py
import os
import shutil
import shortuuid

__version__ = "0.8.0"
__all__ = [
    "PlacementSet",
    "ConnectivitySet",
    "FileStore",
    "MorphologyRepository",
    "HDF5Engine",
    "StorageNode",
]


def on_main(prep=None, ret=None):
    def decorator(f):
        def wrapper(self, *args, **kwargs):
            r = None
            self.comm.barrier()
            if self.comm.get_rank() == 0:
                r = f(self, *args, **kwargs)
            elif prep:
                prep(self, *args, **kwargs)
            self.comm.barrier()
            if not ret:
                return self.comm.bcast(r, root=0)
            else:
                return ret(self, *args, **kwargs)

        return wrapper

    return decorator


def on_main_until(until, prep=None, ret=None):
    def decorator(f):
        def wrapper(self, *args, **kwargs):
            global _procpass
            r = None
            self.comm.barrier()
            if self.comm.get_rank() == 0:
                r = f(self, *args, **kwargs)
            elif prep:
                prep(self, *args, **kwargs)
            self.comm.barrier()
            while not until(self, *args, **kwargs):
                pass
            if not ret:
                return self.comm.bcast(r, root=0)
            else:
                return ret(self, *args, **kwargs)

        return wrapper

    return decorator


def _set_root(self, root):
    self._root = root


def _set_active_cfg(self, config):
    config._meta["active_config"] = True


class NoopLock:
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class HDF5Engine(Engine):
    def __init__(self, root, comm):
        super().__init__(root, comm)
        self._lock = MPILock.sync()
        self._readonly = False

    @property
    def root_slug(self):
        return os.path.relpath(self._root)

    @classmethod
    def recognizes(cls, root):
        try:
            h5py.File(root, "r").close()
            return True
        except Exception:
            return False

    def _read(self):
        if self._readonly:
            return NoopLock()
        else:
            return self._lock.read()

    def _write(self):
        if self._readonly:
            raise IOError("Can't perform write operations in readonly mode.")
        else:
            return self._lock.write()

    def _master_write(self):
        if self._readonly:
            raise IOError("Can't perform write operations in readonly mode.")
        else:
            return self._lock.single_write()

    def _handle(self, mode):
        if self._readonly and mode != "r":
            raise IOError("Can't perform write operations in readonly mode.")
        else:
            return h5py.File(self._root, mode)

    def exists(self):
        return os.path.exists(self._root)

    @on_main_until(lambda self: self.exists())
    def create(self):
        with self._handle("w") as handle:
            handle.create_group("placement")
            handle.create_group("connectivity")
            handle.create_group("files")
            handle.create_group("morphologies")

    @on_main_until(lambda self, r: self.exists(), _set_root)
    def move(self, new_root):
        shutil.move(self._root, new_root)
        self._root = new_root

    @on_main_until(lambda self, r: self.__class__(self.root, self.comm).exists())
    def copy(self, new_root):
        shutil.copy(self._root, new_root)

    @on_main_until(lambda self: not self.exists())
    def remove(self):
        os.remove(self._root)

    @on_main_until(
        lambda self, ct: PlacementSet.exists(self, ct),
        prep=None,
        ret=lambda self, ct: PlacementSet(self, ct),
    )
    def require_placement_set(self, ct):
        return PlacementSet.require(self, ct)

    @on_main()
    def clear_placement(self):
        with self._handle("a") as handle:
            handle.require_group("placement")
            del handle["placement"]
            handle.require_group("placement")
            self._write_chunk_stats(handle, {})

    @on_main()
    def clear_connectivity(self):
        with self._handle("a") as handle:
            handle.require_group("connectivity")
            del handle["connectivity"]
            handle.require_group("connectivity")
            stats = self._read_chunk_stats(handle)
            stats = {
                k: {"placed": v["placed"], "connections": {"inc": 0, "out": 0}}
                for k, v in stats.items()
            }
            self._write_chunk_stats(handle, stats)

    def get_chunk_stats(self):
        with self._handle("r") as handle:
            return self._read_chunk_stats(handle)

    def _read_chunk_stats(self, handle: h5py.File) -> object:
        return json.loads(handle.attrs.get("chunks", "{}"))

    def _write_chunk_stats(self, handle, stats):
        handle.attrs["chunks"] = json.dumps(stats)


def _get_default_root():
    return os.path.abspath(
        os.path.join(
            ".",
            "scaffold_network_"
            + datetime.now().strftime("%Y_%m_%d")
            + "_"
            + shortuuid.uuid()
            + ".hdf5",
        )
    )


@config.node
class StorageNode(IStorageNode):
    root = config.attr(type=str, default=_get_default_root, call_default=True)
    """
    Path to the HDF5 network storage file.
    """



import os, sys, unittest, threading
import numpy as np

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "..", ".."))


def get_data_path(*paths):
    return os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            "..",
            "data",
            *paths,
        )
    )


def get_config_path(file):
    return get_data_path(
        "configs",
        file + (".json" if not file.endswith(".json") else ""),
    )


def get_morphology_path(file):
    return get_data_path(
        "morphologies",
        file,
    )



"""
The chunks module provides the tools for the HDF5 engine to store the chunked placement
data received from the placement module in seperate datasets to arbitrarily parallelize
and scale scaffold models.

The module provides the :class:`.ChunkLoader` mixin for
:class:`~.resource.Resource` objects (e.g. PlacementSet,
ConnectivitySet) to organize :class:`.ChunkedProperty` and :class:`.ChunkedCollection`
objects within them.
"""

from bsb.storage import Chunk, chunklist
import numpy as np
import contextlib
from .resource import handles_handles, HANDLED


class ChunkLoader:
    """
    :class:`~.resource.Resource` mixin to organize chunked properties
    and collections within itself.

    :param properties: An iterable of functions that construct :class:`.ChunkedProperty`.
    :type: Iterable
    :param properties: An iterable of names for constructing :class:`.ChunkedCollection`.
    :type: Iterable
    """

    def __init_subclass__(cls, properties=(), collections=(), **kwargs):
        super().__init_subclass__(**kwargs)
        cls._properties = list(properties)
        cls._collections = list(collections)

    def __init__(self):
        self._chunks = []
        self._properties = []
        self._collections = []
        for prop_constr in self.__class__._properties:
            prop = prop_constr(self)
            self.__dict__[f"_{prop.name}_chunks"] = prop
            self._properties.append(prop)
        for prop_constr in self.__class__._collections:
            prop = prop_constr(self)
            self.__dict__[f"_{prop.collection}_chunks"] = prop
            self._collections.append(prop)

    def get_loaded_chunks(self):
        if not self._chunks:
            return self.get_all_chunks()
        else:
            return self._chunks.copy()

    @handles_handles("r")
    def get_all_chunks(self, handle=HANDLED):
        chunks = list(handle[self._path].keys())
        size = None
        if chunks:
            # If any chunks have been written, this HDF5 file is tagged with a
            # chunk size
            size = self._get_chunk_size(handle)

        return chunklist(Chunk.from_id(int(c), size) for c in chunks)

    @contextlib.contextmanager
    def chunk_context(self, chunks):
        old_chunks = self._chunks
        self._chunks = chunklist(chunks)
        yield
        self._chunks = old_chunks

    def get_chunk_path(self, chunk=None, collection=None, key=None):
        """
        Return the full HDF5 path of a chunk.

        :param chunk: Chunk
        :type chunk: :class:`bsb.storage.Chunk`
        :returns: HDF5 path
        :rtype: str
        """
        path = f"{self._path}"
        if chunk is not None:
            path += f"/{chunk.id}"
            if collection is not None:
                path += f"/{collection}"
            if key is not None:
                path += f"/{key}"
        return path

    def include_chunk(self, chunk):
        """
        Include a chunk in the data when loading properties/collections.
        """
        self._chunks.append(chunk if isinstance(chunk, Chunk) else Chunk(chunk, None))
        self._chunks.sort()

    def exclude_chunk(self, chunk):
        """
        Exclude a chunk from the data when loading properties/collections.
        """
        self._chunks.remove(chunk if isinstance(chunk, Chunk) else Chunk(chunk, None))

    def set_chunk_filter(self, chunks):
        self._chunks = chunklist(chunks)

    def clear_chunk_filter(self):
        self._chunks = []

    @handles_handles("a")
    def require_chunk(self, chunk, handle=HANDLED):
        """
        Create a chunk if it doesn't exist yet, or do nothing.
        """
        path = self.get_chunk_path(chunk)
        if path not in handle:
            chunk_group = handle.create_group(path)
            self._set_chunk_size(handle, chunk.dimensions)
            for p in self._properties:
                chunk_group.create_dataset(
                    f"{path}/{p.name}", p.shape, maxshape=p.maxshape, dtype=p.dtype
                )
            for c in self._collections:
                chunk_group.create_group(path + f"/{c.collection}")

    def clear(self, chunks=None):
        if chunks is None:
            chunks = self.get_loaded_chunks()
        for chunk in chunks:
            for prop in self._properties:
                prop.clear(chunk)
            for coll in self._collections:
                coll.clear(chunk)

    def _set_chunk_size(self, handle, size):
        fsize = handle.attrs.get("chunk_size", np.full(3, np.nan))
        if np.all(np.isnan(fsize)):
            handle.attrs["chunk_size"] = size
        elif not np.all(np.isnan(size)) and not np.allclose(fsize, size):
            raise Exception(f"Chunk size mismatch. File: {fsize}. Given: {size}")

    def _get_chunk_size(self, handle):
        return handle.attrs["chunk_size"]


# The ChunkedProperty and ChunkedCollection are a bit fucked in terms of inheritance and
# how they handle their polymorphism...
class ChunkedProperty:
    """
    Chunked properties are stored inside the ``chunks`` group of the :class:`.ChunkLoader`
    they belong to. Inside the ``chunks`` group another group is created per chunk, inside
    of which a dataset exists per property.
    """

    def __init__(
        self, loader, property, shape, dtype, insert=None, extract=None, collection=None
    ):
        self.loader = loader
        self.name = property
        self.collection = collection
        self.dtype = dtype
        self.shape = shape
        self.insert = insert
        self.extract = extract
        if shape is not None:
            maxshape = list(shape)
            maxshape[0] = None
            self.maxshape = tuple(maxshape)
        else:
            self.maxshape = None

    @handles_handles("r", lambda self: self.loader._engine)
    def load(self, raw=False, key=None, pad_by=None, handle=HANDLED):
        chunks = self.loader.get_loaded_chunks()
        reader = self.get_chunk_reader(handle, raw, key, pad_by=pad_by)
        # Read and collect all non empty chunks
        chunked_data = tuple(data for c in chunks if (data := reader(c)).size)
        # No data? Return empty
        if not chunked_data:
            data = np.empty(self.shape if self.shape is not None else 0)
            if self.extract and not raw:
                return self.extract(data, None)
            else:
                return data
        # Allow custom ndarrays with concatenate methods to concatenate themselves.
        if concatenator := getattr(chunked_data[0].__class__, "concatenate", None):
            return concatenator(*chunked_data)
        else:
            return np.concatenate(chunked_data)

    def get_chunk_reader(self, handle, raw, key=None, pad_by=None):
        """
        Create a chunk reader that either returns the raw data or extracts it.
        """
        key = self.name if key is None else key

        def read_chunk(chunk, pad=0):
            pad_shape = self.shape[1:] if self.shape is not None else ()
            if pad_by:
                pad = len(handle[self._chunk_path(chunk, pad_by)])
            try:
                chunk_group = handle[self._chunk_path(chunk, key)]
            except KeyError:
                chunk_group = None
                data = np.zeros((pad, *pad_shape), dtype=self.dtype)
            else:
                data = chunk_group[()]
                if len(data) < pad:
                    fillshape = (pad - len(data), *pad_shape)
                    data = np.concatenate((data, np.zeros(fillshape, dtype=self.dtype)))
            if not (raw or self.extract is None):
                data = self.extract(data, chunk_group)
            return data

        # Return the created function
        return read_chunk

    @handles_handles("a", lambda self: self.loader._engine)
    def append(self, chunk, data, key=None, handle=HANDLED):
        """
        Append data to a property chunk. Will create it if it doesn't exist.

        :param data: Data to append to the chunked property.
        :param chunk: Chunk
        :type chunk: :class:`bsb.storage.Chunk`
        """
        key = key or self.name
        if self.insert is not None:
            data = self.insert(data)
        self.loader.require_chunk(chunk, handle)
        chunk_group = handle[self._chunk_path(chunk)]
        if key not in chunk_group:
            if self.shape is None:
                shape = data.shape
                maxshape = list(shape)
                maxshape[0] = None
            else:
                shape = list(self.shape)
                shape[0] = len(data)
                maxshape = self.maxshape
            chunk_group.create_dataset(
                key,
                shape,
                data=data,
                maxshape=maxshape,
                dtype=self.dtype,
            )
        else:
            dset = chunk_group[key]
            start_pos = dset.shape[0]
            dset.resize(start_pos + len(data), axis=0)
            dset[start_pos:] = data

    @handles_handles("a", lambda self: self.loader._engine)
    def clear(self, chunk, key=None, handle=HANDLED):
        key = key or self.name
        chunk_group = handle[self._chunk_path(chunk)]
        if key not in chunk_group:
            chunk_group.create_dataset(
                key,
                self.shape,
                data=np.empty(self.shape, dtype=self.dtype),
                maxshape=self.maxshape,
                dtype=self.dtype,
            )
        else:
            dset = chunk_group[key]
            dset.resize(0, axis=0)

    @handles_handles("a", lambda self: self.loader._engine)
    def overwrite(self, chunk, data, key=None, handle=HANDLED):
        self.clear(chunk, key, handle)
        self.append(chunk, data, key, handle)

    def _chunk_path(self, chunk, key=None):
        return self.loader.get_chunk_path(chunk, self.collection, key)


class ChunkedCollection(ChunkedProperty):
    """
    Chunked collections are stored inside the ``chunks`` group of the
    :class:`.ChunkLoader` they belong to. Inside the ``chunks`` group another group is
    created per chunk, inside of which a group exists per collection. Arbitrarily named
    datasets can be stored inside of this collection.
    """

    def __init__(self, loader, collection, shape, dtype, insert=None, extract=None):
        super().__init__(loader, None, shape, dtype, insert, extract, collection)

    @handles_handles("r", lambda self: self.loader._engine)
    def keys(self, handle=HANDLED):
        try:
            return list(handle[self.loader._path].attrs[f"{self.collection}_keys"])
        except KeyError:
            return []

    @handles_handles("r", lambda self: self.loader._engine)
    def load(self, key, handle=HANDLED, **kwargs):
        return super().load(key=key, handle=handle, **kwargs)

    @handles_handles("a", lambda self: self.loader._engine)
    def append(self, chunk, key, data, handle=HANDLED, **kwargs):
        self._add_key(key, handle=handle)
        return super().append(chunk, data, key=key, handle=handle, **kwargs)

    @handles_handles("a", lambda self: self.loader._engine)
    def overwrite(self, chunk, data, key, handle=HANDLED, **kwargs):
        self._add_key(key, handle=handle)
        return super().overwrite(chunk, data, key=key, handle=handle, **kwargs)

    @handles_handles("a", lambda self: self.loader._engine)
    def clear(self, chunk, handle=HANDLED):
        del handle[self._chunk_path(chunk)]
        handle.create_group(self._chunk_path(chunk))

    @handles_handles("a", lambda self: self.loader._engine)
    def _add_key(self, key, handle=HANDLED):
        keys = set(self.keys(handle=handle))
        keys.add(key)
        handle[self.loader._path].attrs[f"{self.collection}_keys"] = list(keys)

    @handles_handles("r", lambda self: self.loader._engine)
    def load_all(self, handle=HANDLED, **kwargs):
        print("?", self)
        return {
            key: super(type(self), self).load(key=key, handle=handle, **kwargs)
            for key in self.keys()
        }



from bsb.exceptions import (
    MissingMorphologyError,
    DatasetExistsError,
    DatasetNotFoundError,
)
from bsb import config
from bsb.storage._chunks import Chunk, chunklist
from bsb._encoding import EncodedLabels
from bsb.storage.interfaces import PlacementSet as IPlacementSet
from bsb.morphologies import MorphologySet, RotationSet
from bsb.morphologies.selector import MorphologySelector
from .resource import Resource, handles_handles, HANDLED
from .chunks import ChunkLoader, ChunkedProperty, ChunkedCollection
import numpy as np
import itertools
import json

_root = "/placement/"


@config.node
class _MapSelector(MorphologySelector):
    ps = config.attr(type=lambda x: x)
    names = config.attr(type=lambda x: x)

    def __init__(self, *, ps=None, names=None):
        self._ps = ps
        self._names = set(names)

    def validate(self, loaders):
        missing = set(self._names) - {m.get_meta()["name"] for m in loaders}
        if missing:
            raise MissingMorphologyError(
                "Morphology repository misses the following morphologies required by"
                + f" {self._ps.tag}: {', '.join(missing)}"
            )

    def pick(self, stored_morphology):
        name = stored_morphology.get_meta()["name"]
        return name in self._names


_ps_properties = (
    lambda loader: ChunkedProperty(loader, "position", shape=(0, 3), dtype=float),
    lambda loader: ChunkedProperty(loader, "rotation", shape=(0, 3), dtype=float),
    lambda loader: ChunkedProperty(loader, "morphology", shape=(0,), dtype=int),
    lambda loader: ChunkedProperty(
        loader, "labels", shape=(0,), dtype=int, extract=encode_labels
    ),
)
_ps_collections = (
    lambda loader: ChunkedCollection(loader, "additional", shape=None, dtype=float),
)


class PlacementSet(
    Resource,
    ChunkLoader,
    IPlacementSet,
    properties=_ps_properties,
    collections=_ps_collections,
):
    """
    Fetches placement data from storage.

    .. note::

        Use :meth:`Scaffold.get_placement_set <bsb.core.Scaffold.get_placement_set>` to
        correctly obtain a PlacementSet.
    """

    _position_chunks: ChunkedProperty
    _morphology_chunks: ChunkedProperty
    _rotation_chunks: ChunkedProperty
    _labels_chunks: ChunkedProperty
    _additional_chunks: ChunkedCollection

    def __init__(self, engine, cell_type):
        tag = cell_type.name
        Resource.__init__(self, engine, _root + tag)
        IPlacementSet.__init__(self, engine, cell_type)
        ChunkLoader.__init__(self)
        self._labels = None
        self._morphology_labels = None
        if not self.exists(engine, cell_type):
            raise DatasetNotFoundError(f"PlacementSet '{tag}' does not exist")

    @classmethod
    def create(cls, engine, cell_type):
        """
        Create the structure for this placement set in the HDF5 file. Placement sets are
        stored under ``/placement/<tag>``.
        """
        tag = cell_type.name
        path = _root + tag
        with engine._write():
            with engine._handle("a") as h:
                if path in h:
                    raise DatasetExistsError(f"PlacementSet '{tag}' already exists.")
                h.create_group(path)
        return cls(engine, cell_type)

    @staticmethod
    def exists(engine, cell_type):
        with engine._read():
            with engine._handle("r") as h:
                return "/placement/" + cell_type.name in h

    @classmethod
    def require(cls, engine, cell_type):
        tag = cell_type.name
        path = _root + tag
        with engine._write():
            with engine._handle("a") as h:
                h.require_group(path)
        return cls(engine, cell_type)

    @handles_handles("r")
    def load_positions(self, handle=HANDLED):
        """
        Load the cell positions.

        :raises: DatasetNotFoundError when there is no rotation information for this
           cell type.
        """
        try:
            positions = self._position_chunks.load(handle=handle)
        except DatasetNotFoundError:
            raise DatasetNotFoundError(
                f"No position information for the '{self.tag}' placement set."
            )
        else:
            if self._labels:
                mask = self.get_label_mask(self._labels, handle=handle)
                return positions[mask]
            else:
                return positions

    @handles_handles("r")
    def load_rotations(self, handle=HANDLED):
        """
        Load the cell rotations.

        :raises: DatasetNotFoundError when there is no rotation information for this
           cell type.
        """
        data = self._rotation_chunks.load(handle=handle)
        if len(data) == 0 and len(self) != 0:
            raise DatasetNotFoundError("No rotation data available.")
        if self._labels:
            mask = self.get_label_mask(self._labels, handle=handle)
            data = data[mask]
        return RotationSet(data)

    @handles_handles("r")
    def load_morphologies(self, handle=HANDLED, allow_empty=False):
        """
        Load the cell morphologies.

        :raises: DatasetNotFoundError when the morphology data is not found.
        """
        reader = self._morphology_chunks.get_chunk_reader(handle, True)
        loaders = self._get_morphology_loaders(handle=handle)
        data = []
        for chunk in self.get_loaded_chunks():
            path = self.get_chunk_path(chunk)
            try:
                _map = handle[path].attrs["morphology_loaders"]
            except KeyError:
                continue
            block = np.vectorize(list(loaders.keys()).index)(_map[reader(chunk)])
            if len(block):
                data.append(block)
        if len(data) == 0 and (len(self) != 0 or len(loaders) == 0):
            if not allow_empty:
                raise DatasetNotFoundError("No morphology data available.")
            else:
                data = np.empty(0, dtype=int)
        else:
            data = np.concatenate(data)
        if self._labels:
            mask = self.get_label_mask(self._labels, handle=handle)
            data = data[mask]
        return MorphologySet(
            list(loaders.values()),
            data,
            labels=self._morphology_labels,
        )

    @handles_handles("r")
    def load_additional(self, key=None, handle=HANDLED):
        if key is None:
            return self._additional_chunks.load_all()
        else:
            return self._additional_chunks.load(key)

    @handles_handles("r")
    def _get_morphology_loaders(self, handle=HANDLED):
        stor_mor = {}
        for chunk in self.get_loaded_chunks():
            path = self.get_chunk_path(chunk)
            try:
                _map = handle[path].attrs["morphology_loaders"]
            except KeyError:
                continue
            stor_mor.update(
                (m.name, m)
                for m in self._engine.morphologies.select(
                    _MapSelector(ps=self, names=_map)
                )
            )
        return stor_mor

    @handles_handles("a")
    def _set_morphology_loaders(self, map, handle=HANDLED):
        for chunk in self.get_loaded_chunks():
            path = self.get_chunk_path(chunk)
            handle[path].attrs["morphology_loaders"] = map

    def __iter__(self):
        return itertools.zip_longest(
            self.load_positions(),
            self.load_morphologies(),
        )

    def __len__(self):
        if self._labels:
            return np.sum(self._labels_chunks.load().get_mask(self._labels))
        else:
            return len(self._position_chunks.load())

    @handles_handles("a")
    def append_data(
        self,
        chunk,
        positions=None,
        morphologies=None,
        rotations=None,
        additional=None,
        count=None,
        handle=HANDLED,
    ):
        """
        Append data to the placement set.

        :param chunk: The chunk to store data in.
        :param positions: Cell positions
        :type positions: :class:`numpy.ndarray`
        :param rotations: Cell rotations
        :type rotations: ~bsb.morphologies.RotationSet
        :param morphologies: Cell morphologies
        :type morphologies: ~bsb.morphologies.MorphologySet
        :param count: Amount of entities to place. Excludes the use of any positional,
          rotational or morphological data.
        :type count: int
        """
        if not isinstance(chunk, Chunk):
            chunk = Chunk(chunk, None)
        if positions is not None:
            positions = np.array(positions, copy=False)
        if count is not None:
            if not (positions is None and morphologies is None):
                raise ValueError(
                    "The `count` keyword is reserved for creating entities,"
                    + " without any positional, or morphological data."
                )
            self.require_chunk(chunk, handle=handle)

        if positions is not None:
            self._position_chunks.append(chunk, positions)
        if morphologies is not None:
            self._append_morphologies(chunk, morphologies)
            if rotations is None:
                rotations = np.zeros((len(morphologies), 3))
        if rotations is not None:
            self._rotation_chunks.append(chunk, rotations)

        if additional is not None:
            for key, ds in additional.items():
                self.append_additional(key, chunk, ds)
        self._track_add(handle, chunk, len(positions))

    def _append_morphologies(self, chunk, new_set):
        with self.chunk_context([chunk]):
            morphology_set = self.load_morphologies(allow_empty=True).merge(new_set)
            self._set_morphology_loaders(morphology_set._serialize_loaders())
            self._morphology_chunks.clear(chunk)
            self._morphology_chunks.append(chunk, morphology_set.get_indices())

    def append_entities(self, chunk, count, additional=None):
        self.append_data(chunk, count=count, additional=additional)

    def append_additional(self, name, chunk, data):
        self._additional_chunks.append(chunk, name, data)

    @handles_handles("a")
    def label_by_mask(self, mask, labels, handle=HANDLED):
        cells = np.array(mask, copy=False)
        if cells.dtype != bool or len(cells) != len(self):
            raise LabellingException("Mask doesn't fit data.")
        self._write_labels(
            labels,
            handle,
            lambda: self._lendemux(),
            lambda s: cells[s],
        )

    @handles_handles("a")
    def label(self, labels, cells, handle=HANDLED):
        cells = np.array(cells, copy=False)
        len_ = len(self)
        oob = cells > len_
        if np.any(oob):
            oob_idx = cells[oob]
            raise LabellingException(
                f"Cell labels {oob_idx} out of range for placement set with size {len_}."
            )
        self._write_labels(
            labels,
            handle,
            lambda: self._demux(cells),
            lambda x: x,
        )

    def _write_labels(self, labels, handle, demux_f, data_f):
        # Create a label reader that can read out label data per chunk, and pads missing
        # cells with unlabelled cells.
        label_reader = self._labels_chunks.get_chunk_reader(
            handle, False, pad_by="position"
        )
        updated_labels = None
        # Demultiplex the cells per chunk. demux_f sets chunk context to current chunk
        for chunk, block in demux_f():
            enc_labels = label_reader(chunk)
            # The label reader gets the labelsets from a shared attribute on the PS, so we
            # keep and update 1 shared reference, that we update after the loop.
            if updated_labels:
                enc_labels.labels = updated_labels
            else:
                updated_labels = enc_labels.labels
            # Label the cells
            enc_labels.label(labels, data_f(block))
            # Overwrite with new labelled data.
            self._labels_chunks.overwrite(chunk, enc_labels, handle=handle)
        # Update the shared labelset reference on the PS.
        if updated_labels is not None:
            handle[self._path].attrs["labelsets"] = json.dumps(
                updated_labels, default=list
            )

    def set_label_filter(self, labels):
        self._labels = labels

    def set_morphology_label_filter(self, morphology_labels):
        """
        Sets the labels by which any morphology loaded from this set will be filtered.

        :param morphology_labels: List of labels to filter the morphologies by.
        :type morphology_labels: List[str]
        """
        self._morphology_labels = morphology_labels

    @handles_handles("r")
    def get_labelled(self, labels, handle=HANDLED):
        mask = self.get_label_mask(labels, handle=handle)
        return np.nonzero(mask)[0]

    @handles_handles("r")
    def get_label_mask(self, labels, handle=HANDLED):
        return self._labels_chunks.load(handle=handle, pad_by="position").get_mask(labels)

    def _lendemux(self):
        """
        .. warning::

            This function sets the chunk context for as long as it iterates.
        """
        ctr = 0
        for chunk in self.get_loaded_chunks():
            with self.chunk_context([chunk]):
                len_ = len(self)
                yield chunk, slice(ctr, ctr := ctr + len_)

    def _demux(self, ids):
        """
        .. warning::

            This function sets the chunk context for as long as it iterates.
        """
        for chunk in self.get_loaded_chunks():
            with self.chunk_context([chunk]):
                ln = len(self)
                idx = ids < ln
                block = ids[idx]
                yield chunk, block
                ids = ids[~idx]
                ids -= ln

    def _track_add(self, handle, chunk, count):
        # Track addition in global chunk stats
        global_stats = json.loads(handle.attrs.get("chunks", "{}"))
        stats = global_stats.setdefault(
            str(chunk.id), {"placed": 0, "connections": {"inc": 0, "out": 0}}
        )
        stats["placed"] += count
        handle.attrs["chunks"] = json.dumps(global_stats)
        # Track addition in placement set
        handle[self._path].attrs["len"] = handle[self._path].attrs.get("len", 0) + count
        chunk_stats = json.loads(handle[self._path].attrs.get("chunks", "{}"))
        chunk_stats[str(chunk.id)] = chunk_stats.get(str(chunk.id), 0) + count
        handle[self._path].attrs["chunks"] = json.dumps(chunk_stats)

    @handles_handles("r")
    def get_chunk_stats(self, handle=HANDLED):
        return json.loads(handle[self._path].attrs["chunks"])

    @handles_handles("r")
    def load_ids(self, handle=HANDLED):
        if self._chunks is None:
            return np.arange(len(self))
        stats = self.get_chunk_stats(handle)
        offsets = {}
        ctr = 0
        return np.concatenate(
            [
                np.arange(ctr, (ctr := ctr + len_))
                for chunk, len_ in sorted(
                    stats.items(), key=lambda k: Chunk.from_id(int(k[0]), None).id
                )
                if chunk in self._chunks
            ]
        )


def encode_labels(data, ds):
    if ds is None:
        return EncodedLabels.none(len(data))
    ps_group = ds.parent.parent
    serialized = json.dumps(EncodedLabels.none(1).labels, default=list)
    labels = json.loads(ps_group.attrs.get("labelsets", serialized))
    return EncodedLabels(shape=data.shape, buffer=data, labels=labels)


class LabellingException(Exception):
    pass



from bsb.services import MPI
from bsb.storage import Storage
import numpy as np
import unittest
import json
import h5py
import os


class TestHandcrafted(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        if MPI.get_rank() == 0:
            with h5py.File("test.h5", "w") as f:
                g = f.create_group("morphologies")
                g = g.create_group("M")
                ds = g.create_dataset("data", data=np.empty((0, 5)))
                ds.attrs["labels"] = json.dumps({0: []})
                ds.attrs["properties"] = []
                g.create_dataset("graph", data=[])
                f.create_dataset("morphology_meta", data=json.dumps({"M": {}}))
            with h5py.File("test2.h5", "w") as f:
                g = f.create_group("morphologies")
                g = g.create_group("M")
                ds = g.create_dataset("data", data=np.empty((0, 5)))
                ds.attrs["labels"] = json.dumps({0: []})
                ds.attrs["properties"] = []
                g.create_dataset("graph", data=[[0, -1], [0, -1], [0, -1]])
                f.create_dataset("morphology_meta", data=json.dumps({"M": {}}))
            with h5py.File("test3.h5", "w") as f:
                g = f.create_group("morphologies")
                g = g.create_group("M")
                ds = g.create_dataset("data", data=np.ones((1, 5)))
                ds.attrs["labels"] = json.dumps({1: []})
                ds.attrs["properties"] = []
                g.create_dataset("graph", data=[[0, -1]])
                f.create_dataset("morphology_meta", data=json.dumps({"M": {}}))
            with h5py.File("test4.h5", "w") as f:
                g = f.create_group("morphologies")
                g = g.create_group("M")
                data = np.ones((5, 5))
                data[:, 0] = np.arange(5) * 2
                data[:, 1] = np.arange(5) * 2
                data[:, 2] = np.arange(5) * 2
                data[:, 3] = np.arange(5) * 2
                ds = g.create_dataset("data", data=data)
                ds.attrs["labels"] = json.dumps({1: []})
                ds.attrs["properties"] = []
                g.create_dataset("graph", data=[[i + 1, -1] for i in range(5)])
                f.create_dataset("morphology_meta", data=json.dumps({"M": {}}))
            with h5py.File("test5.h5", "w") as f:
                g = f.create_group("morphologies")
                g = g.create_group("M")
                data = np.ones((5, 5))
                data[:, 0] = np.arange(5) * 2
                data[:, 1] = np.arange(5) * 2
                data[:, 2] = np.arange(5) * 2
                data[:, 3] = np.arange(5) * 2
                ds = g.create_dataset("data", data=data)
                ds.attrs["labels"] = json.dumps({1: []})
                ds.attrs["properties"] = []
                g.create_dataset("graph", data=[[i + 1, -1] for i in range(4)] + [[5, 0]])
                f.create_dataset("morphology_meta", data=json.dumps({"M": {}}))
        MPI.barrier()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        if MPI.get_rank() == 0:
            os.remove("test.h5")
            os.remove("test2.h5")
            os.remove("test3.h5")
            os.remove("test4.h5")
            os.remove("test5.h5")
        MPI.barrier()

    def test_empty_repository(self):
        pass

    def test_empty(self):
        mr = Storage("hdf5", "test.h5").morphologies
        m = mr.load("M")
        msg = "Empty morfo should not have root branches"
        self.assertEqual(0, len(m.roots), msg)
        msg = "Empty morfo should not have branches"
        self.assertEqual(0, len(m.branches), msg)
        msg = "Empty morfo should not have points"
        self.assertEqual(0, len(m.flatten()), msg)
        self.assertEqual(0, len(m), msg)
        self.assertTrue(m._check_shared(), "Empty morpho not shared")

    def test_empty_branches(self):
        mr = Storage("hdf5", "test2.h5").morphologies
        m = mr.load("M")
        msg = "Empty unattached branches should still be root."
        self.assertEqual(3, len(m.roots), msg)
        self.assertEqual(3, len(m.branches), "Missing branch")
        msg = "Empty morfo should not have points, even when it has empty branches"
        self.assertEqual(0, len(m), msg)
        self.assertEqual(0, len(m.flatten()), msg)
        self.assertTrue(m._check_shared(), "Load should produce shared")

    def test_single_branch_single_element(self):
        mr = Storage("hdf5", "test3.h5").morphologies
        m = mr.load("M")
        msg = "Single point unattached branches should still be root."
        self.assertEqual(1, len(m.roots), msg)
        self.assertEqual(1, len(m.branches), "Missing branch")
        msg = "Flatten of single point should produce 1 x 3 matrix."
        self.assertEqual((1, 3), m.flatten().shape, msg)
        msg = "should produce 1 element vector."
        self.assertEqual((1,), m.flatten_radii().shape, msg)
        self.assertEqual((1,), m.flatten_labels().shape, msg)
        msg = "Flatten without properties should produce n x 0 matrix."
        self.assertEqual({}, m.flatten_properties(), msg)

    def test_multi_branch_single_element(self):
        mr = Storage("hdf5", "test4.h5").morphologies
        m = mr.load("M")
        msg = "Single point unattached branches should still be root."
        self.assertEqual(5, len(m.roots), msg)
        self.assertEqual(5, len(m.branches), "Missing branch")
        msg = (
            "Flatten of single point branches should produce n-branch x n-vectors matrix."
        )
        matrix = m.flatten()
        self.assertEqual((5, 3), matrix.shape, msg)
        msg = "Flatten produced an incorrect matrix"
        self.assertTrue(
            np.array_equal(np.array([[i * 2] * 3 for i in range(5)]), matrix), msg
        )

    def test_multi_branch_single_element_depth_first(self):
        mr = Storage("hdf5", "test5.h5").morphologies
        m = mr.load("M")
        msg = "1 out of 5 branches was attached, 4 roots expected."
        self.assertEqual(4, len(m.roots), msg)
        self.assertEqual(5, len(m.branches), "Missing branch")



from bsb.unittest.engines import (
    TestStorage as _TestStorage,
    TestPlacementSet as _TestPlacementSet,
    TestMorphologyRepository as _TestMorphologyRepository,
    TestConnectivitySet as _TestConnectivitySet,
)
import unittest


class TestStorage(_TestStorage, unittest.TestCase, engine_name="hdf5"):
    pass


class TestPlacementSet(_TestPlacementSet, unittest.TestCase, engine_name="hdf5"):
    pass


class TestMorphologyRepository(
    _TestMorphologyRepository, unittest.TestCase, engine_name="hdf5"
):
    pass


class TestConnectivitySet(_TestConnectivitySet, unittest.TestCase, engine_name="hdf5"):
    pass



import setuptools
import os

_findver = "__version__ = "
_rootpath = os.path.join(os.path.dirname(__file__), "bsb_hdf5", "__init__.py")
with open(_rootpath, "r") as f:
    for line in f:
        if "__version__ = " in line:
            f = line.find(_findver)
            __version__ = eval(line[line.find(_findver) + len(_findver) :])
            break
    else:
        raise Exception(f"No `__version__` found in '{_rootpath}'.")

with open("README.md", "r") as fh:
    long_description = fh.read()

requires = [
    "bsb~=4.0.0a54",
    "shortuuid",
]

setuptools.setup(
    name="bsb-hdf5",
    version=__version__,
    author="Robin De Schepper, Egidio D'Angelo, Claudia Casellato",
    author_email="robingilbert.deschepper@unipv.it",
    description="The HDF5 storage engine of the BSB",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dbbs-lab/bsb-hdf5",
    license="GPLv3",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "bsb.storage.engines": ["hdf5 = bsb_hdf5"],
    },
    python_requires="~=3.8",
    install_requires=requires,
    project_urls={
        "Bug Tracker": "https://github.com/dbbs-lab/bsb-hdf5/issues/",
        "Documentation": "https://bsb-hdf5.readthedocs.io/",
        "Source Code": "https://github.com/dbbs-lab/bsb-hdf5/",
    },
    extras_require={"dev": ["sphinx", "sphinxemoji", "furo", "sphinx_design"]},
)



import typing

import numpy as np
import h5py
import inspect
import functools

if typing.TYPE_CHECKING:
    from . import HDF5Engine

# Semantic marker for things that get injected
HANDLED = None


# Decorator to inject handles
def handles_handles(handle_type, handler=lambda self: self._engine):
    lock_f = {"r": lambda eng: eng._read, "a": lambda eng: eng._write}.get(handle_type)

    def decorator(f):
        sig = inspect.signature(f)
        if "handle" not in sig.parameters:
            raise ValueError(
                f"`{f.__module__}.{f.__name__}` needs handle to be handled by "
                f"handles_handles. Clearly."
            )

        @functools.wraps(f)
        def handle_indirection(self, *args, handle=None, **kwargs):
            engine = handler(self)
            lock = lock_f(engine)
            try:
                bound = sig.bind(self, *args, **kwargs)
            except TypeError:
                # Re-call the actual function, for better TypeError
                try:
                    f(self, *args, **kwargs)
                except TypeError as e:
                    # Re-raise the exception from None for better stack trace
                    raise e from None
            if bound.arguments.get("handle", None) is None:
                with lock():
                    with engine._handle(handle_type) as handle:
                        bound.arguments["handle"] = handle
                        return f(*bound.args, **bound.kwargs)
            else:
                return f(*bound.args, **bound.kwargs)

        return handle_indirection

    return decorator


class Resource:
    def __init__(self, engine: "HDF5Engine", path: str):
        self._engine: "HDF5Engine" = engine
        self._path = path

    def __eq__(self, other):
        return (
            self._engine == getattr(other, "_engine", None) and self._path == other._path
        )

    def require(self, handle):
        return handle.require_group(self._path)

    def create(self, data, *args, **kwargs):
        with self._engine._write():
            with self._engine._handle("a") as f:
                f.create_dataset(self._path, data=data, *args, **kwargs)

    def keys(self):
        with self._engine._read():
            with self._engine._handle("r") as f:
                node = f[self._path]
                if isinstance(node, h5py.Group):
                    return list(node.keys())

    def remove(self):
        with self._engine._write():
            with self._engine._handle("a") as f:
                del f[self._path]

    def get_dataset(self, selector=()):
        with self._engine._read():
            with self._engine._handle("r") as f:
                return f[self._path][selector]

    @property
    def attributes(self):
        with self._engine._read():
            with self._engine._handle("r") as f:
                return dict(f[self._path].attrs)

    def get_attribute(self, name):
        attrs = self.attributes
        if name not in attrs:
            raise AttributeError(
                "Attribute '{}' not found in '{}'".format(name, self._path)
            )
        return attrs[name]

    def exists(self):
        with self._engine._read():
            with self._engine._handle("r") as f:
                return self._path in f

    def unmap(self, selector=(), mapping=lambda m, x: m[x], data=None):
        if data is None:
            data = self.get_dataset(selector)
        map = self.get_attribute("map")
        unmapped = []
        for record in data:
            unmapped.append(mapping(map, record))
        return np.array(unmapped)

    def unmap_one(self, data, mapping=None):
        if mapping is None:
            return self.unmap(data=[data])
        else:
            return self.unmap(data=[data], mapping=mapping)

    def __iter__(self):
        return iter(self.get_dataset())

    @property
    def shape(self):
        with self._engine._read():
            with self._engine._handle("r") as f:
                return f[self._path].shape

    def __len__(self):
        return self.shape[0]

    def append(self, new_data, dtype=float):
        if type(new_data) is not np.ndarray:
            new_data = np.array(new_data)
        with self._engine._write():
            with self._engine._handle("a") as f:
                try:
                    d = f[self._path]
                except Exception:
                    shape = list(new_data.shape)
                    shape[0] = None
                    d = f.create_dataset(
                        self._path, data=new_data, dtype=dtype, maxshape=tuple(shape)
                    )
                else:
                    len_ = d.shape[0]
                    len_ += len(new_data)
                    d.resize(len_, axis=0)
                    d[-len(new_data) :] = new_data
